//
//  SendableWrapper.swift
//  E-Invoice
//
//  Created by Christoph Vogelbusch on 15.10.24.
//
import Foundation

public enum DateParseError: Error  {
    case unknownFormat(String)
    public var localizedDescription: String {
        switch self {
            case .unknownFormat(let format):
            return "Unknown date format: '\(format)'"
        }
    }
}

extension DateParseError: LocalizedError {
    public var errorDescription: String? {
        localizedDescription
    }
}

public struct XMLDate: Codable, Sendable, Hashable {
    static let dateStrategy = Date.ParseStrategy(format: "\(year: .defaultDigits)-\(month: .twoDigits)-\(day: .twoDigits)", timeZone: .gmt)
    static let dateTime24Strategy = Date.ParseStrategy(format: "\(year: .defaultDigits)-\(month: .twoDigits)-\(day: .twoDigits)'T'\(hour: .twoDigits(clock: .twentyFourHour, hourCycle: .oneBased)):\(minute: .twoDigits):\(second: .twoDigits)Z", timeZone: .gmt)
    static let dateTime12Strategy = Date.ParseStrategy(format: "\(year: .defaultDigits)-\(month: .twoDigits)-\(day: .twoDigits)'T'\(hour: .twoDigits(clock: .twelveHour, hourCycle: .oneBased)):\(minute: .twoDigits):\(second: .twoDigits)Z", timeZone: .gmt)
    public var date: Date
    public init(from decoder: any Decoder) throws {
        let string = try String(from: decoder)
        
        if let aDate = (try? Date(string, strategy: XMLDate.dateStrategy)) ??
            (try? Date(string, strategy: .dateTime)) ??
            (try? Date(string, strategy: .iso8601)) ??
            (try? Date(string, strategy: XMLDate.dateTime24Strategy)) ??
            (try? Date(string, strategy: XMLDate.dateTime12Strategy)) {
            date = aDate
        } else {
            throw DateParseError.unknownFormat(string)
        }
    }
    
    public init(date: Date) {
        self.date = date
    }
        
    public func encode(to encoder: any Encoder) throws {
        let string = date.formatted(.iso8601)
        try string.encode(to: encoder)
    }
    
    var description: String {
        date.description
    }
}

public protocol CreateEmpty {
    static func empty() -> Self
}

@available(macOS 14, iOS 17, tvOS 17, watchOS 10, *)
@dynamicMemberLookup
@Observable public final class SendableWrapper<WrappedValue: Codable>: Sendable where WrappedValue: Sendable {
    nonisolated(unsafe) private var _wrappedValue: WrappedValue
    public var wrappedValue: WrappedValue {
        get {
            if Thread.isMainThread {
                return _wrappedValue
            } else {
                return DispatchQueue.main.sync { _wrappedValue }
            }
        }
        set {
            if Thread.isMainThread {
                _wrappedValue = newValue
            } else {
                DispatchQueue.main.sync { _wrappedValue = newValue }
            }
        }
    }
    
    public init(_ value: WrappedValue) {
        _wrappedValue = value
    }
    
    public subscript<T>(dynamicMember dynamicMember: WritableKeyPath<WrappedValue, T>) -> T {
        get {
            wrappedValue[keyPath: dynamicMember]
        }
        set {
            wrappedValue[keyPath: dynamicMember] = newValue
        }
    }
    
    public subscript<T>(dynamicMember dynamicMember: KeyPath<WrappedValue, T>) -> T {
        wrappedValue[keyPath: dynamicMember]
    }
}

@available(macOS 14, iOS 17, tvOS 17, watchOS 10, *)
extension SendableWrapper: Codable {
    public convenience init(from decoder: any Decoder) throws {
        self.init(try WrappedValue(from: decoder))
    }
    
    public func encode(to encoder: any Encoder) throws {
        try wrappedValue.encode(to: encoder)
    }
}

@available(macOS 14, iOS 17, tvOS 17, watchOS 10, *)
extension SendableWrapper: Hashable, Equatable where WrappedValue: Hashable {
    public static func == (lhs: SendableWrapper<WrappedValue>, rhs: SendableWrapper<WrappedValue>) -> Bool {
        lhs.wrappedValue == rhs.wrappedValue
    }
    
    public func hash(into hasher: inout Hasher) {
        hasher.combine(wrappedValue)
    }
}

@available(macOS 14, iOS 17, tvOS 17, watchOS 10, *)
extension SendableWrapper: CreateEmpty where WrappedValue: CreateEmpty {
    public static func empty() -> SendableWrapper<WrappedValue> {
        SendableWrapper(WrappedValue.empty())
    }
}

