//
//  ViewPresentable.swift
//  E-Invoice
//
//  Created by Christoph Vogelbusch on 15.10.24.
//

import SwiftUI

public protocol AllCases {
    static var allCases: [Self] { get }
}

@available(macOS 10.15, iOS 14, tvOS 14, watchOS 7, *)
public protocol ViewPresentable {
    associatedtype PresentableView: View
    @MainActor func view(showNil: Bool, search: String?) -> PresentableView
}

@available(macOS 10.15, iOS 14, tvOS 14, watchOS 7, *)
public protocol KeyViewPresentable: Searchable, ViewPresentable {
    associatedtype ElementKey: Hashable
    associatedtype SinglePresentableView: View
    static var elementKeys: [ElementKey] { get }
    var hasNilValue: Bool { get }
    func title(for key: ElementKey) -> String
    func details(for key: ElementKey) -> String?
    @MainActor func view(for key: ElementKey, show: Bool, search: String?) -> SinglePresentableView
}

@available(macOS 10.15, iOS 14, tvOS 14, watchOS 7, *)
public extension KeyViewPresentable {
    @MainActor func view(showNil: Bool, search: String?) -> some View {
        ForEach(Self.elementKeys, id: \.self) { key in
            self.view(for: key, show: showNil, search: search)
        }
    }

    @MainActor @ViewBuilder
    func presentableView<T:ViewPresentable>(for array: [SendableWrapper<T>], title: String, details: String?, show: Bool, search: String?) -> some View where T: Hashable, T: KeyViewPresentable {
        if let value = array.first {
            if (array.count == 1) {
                TitledDisclosureView(title: title, details: details, canHide: value.hasNilValue) { showNil in
                    value.wrappedValue.view(showNil: showNil, search: search)
                }
            } else {
                ForEach(array, id: \.self) { presentable in
                    TitledDisclosureView(title: title, details: details, canHide: presentable.hasNilValue) { showNil in
                        presentable.wrappedValue.view(showNil: showNil, search: search)
                    }
                }
            }
        } else if show {
            HStack(alignment: .top) {
                SwiftUI.Text("\(title):")
                    .optionalHelp(details)
                SwiftUI.Text("-")
                    .font(.body.bold())
            }
        }
    }
    
    @MainActor @ViewBuilder
    func presentableView<T:ViewPresentable>(for presentable: SendableWrapper<T>, title: String, details: String?, show: Bool, search: String?) -> some View where T: Hashable, T: KeyViewPresentable {
        TitledDisclosureView(title: title, details: details, canHide: presentable.hasNilValue) { showNil in
            presentable.wrappedValue.view(showNil: showNil, search: search)
        }
    }
    
    @MainActor @ViewBuilder
    func presentableView<T:ViewPresentable>(for presentable: SendableWrapper<T>?, title: String, details: String?, show: Bool, search: String?) -> some View where T: Hashable, T: KeyViewPresentable {
        if let presentable {
            TitledDisclosureView(title: title, details: details, canHide: presentable.hasNilValue) { showNil in
                presentable.wrappedValue.view(showNil: showNil, search: search)
            }
        } else if show {
            HStack(alignment: .top) {
                SwiftUI.Text("\(title):")
                    .optionalHelp(details)
                SwiftUI.Text("-")
                    .font(.body.bold())
            }
        }
    }
    
    @MainActor @ViewBuilder
    func anyView<T>(for array: [T], title: String, details: String?, show: Bool, search: String?) -> some View where T: Hashable {
        if let value = array.first {
            if (array.count == 1) {
                HStack(alignment: .top) {
                    SwiftUI.Text("\(title):")
                        .optionalHelp(details)
                    SwiftUI.Text("\(value)")
                        .font(.body.bold())
                }
            } else {
                TitledDisclosureView(title: title, details: details, canHide: false) { _ in
                    ForEach(array, id: \.self) { value in
                        SwiftUI.Text("\(value)")
                    }
                }
            }
        } else if show {
            HStack(alignment: .top) {
                SwiftUI.Text("\(title):")
                    .optionalHelp(details)
                SwiftUI.Text("-")
                    .font(.body.bold())
            }
        }
    }
    
    @MainActor @ViewBuilder
    func anyView<T>(for value: T, title: String, details: String?, show: Bool, search: String?) -> some View where T: Hashable {
        HStack(alignment: .top) {
            SwiftUI.Text("\(title):")
                .optionalHelp(details)
            SwiftUI.Text("\(value)")
                .font(.body.bold())
        }
    }
    
    @MainActor @ViewBuilder
    func anyView<T>(for value: T?, title: String, details: String?, show: Bool, search: String?) -> some View where T: Hashable {
        if let value {
            HStack(alignment: .top) {
                SwiftUI.Text("\(title):")
                    .optionalHelp(details)
                SwiftUI.Text("\(value)")
                    .font(.body.bold())
            }
        } else if show {
            HStack(alignment: .top) {
                SwiftUI.Text("\(title):")
                    .optionalHelp(details)
                SwiftUI.Text("-")
                    .font(.body.bold())
            }
        }
    }
}

public extension DispatchQueue {
    static func executeOnMain<ReturnType: Sendable>(_ closure: @MainActor () throws -> ReturnType) rethrows -> ReturnType {
        if Thread.isMainThread {
            try MainActor.assumeIsolated(closure)
        } else {
            try DispatchQueue.main.sync(execute: closure)
        }
    }
}

@propertyWrapper
public struct ThreadSafe<Value>: Sendable {
    private let lock = NSLock()
    nonisolated(unsafe) private var value: Value
    
    public init(wrappedValue: Value) {
        self.value = wrappedValue
    }
    
    public var wrappedValue: Value {
        get {
            lock.lock()
            defer { lock.unlock() }
            return value
        }
        set {
            lock.lock()
            defer { lock.unlock() }
            value = newValue
        }
    }
    
    public var projectedValue: ThreadSafe<Value> {
        self
    }
}

extension SendableWrapper: ViewPresentable where WrappedValue: ViewPresentable {
    @MainActor public func view(showNil: Bool, search: String?) -> some View {
        wrappedValue.view(showNil: showNil, search: search)
    }
}

@available(macOS 14, iOS 17, tvOS 17, watchOS 10, *)
extension SendableWrapper: KeyViewPresentable where WrappedValue: KeyViewPresentable {
    public static var elementKeys: [WrappedValue.ElementKey] { WrappedValue.elementKeys }
    public var hasNilValue: Bool { wrappedValue.hasNilValue }
    public func title(for key: WrappedValue.ElementKey) -> String { wrappedValue.title(for: key)}
    public func details(for key: WrappedValue.ElementKey) -> String? { wrappedValue.details(for: key)}
    @MainActor public func view(for key: WrappedValue.ElementKey, show: Bool, search: String?) -> some View  { wrappedValue.view(for: key, show: show, search: search)}
}

public protocol TypeNamable {
    var typeName: String { get } // give back a human understandable title for each enum entry
}

public protocol TypeDescribable: TypeNamable {
    var details: String { get } // give back a sussinc explaing description for each enum entry
}

public enum SearchResult: Equatable {
    case title(String), content(String)
}

public protocol Searchable {
    func search(_ string: String) -> SearchResult?
}

extension String: Searchable {
    public func search(_ string: String) -> SearchResult? {
        guard let _ = range(of: string, options: .caseInsensitive) else { return nil }
        return .content(self)
    }
}

extension XMLDate: Searchable {
    public func search(_ string: String) -> SearchResult? {
        guard "\(date)".contains(string) else { return nil }
        return .content(description)
    }
}

extension Decimal: Searchable {
    public func search(_ string: String) -> SearchResult? {
        guard "\(self)".contains(string) else { return nil }
        return .content(description)
    }
}

extension Bool: Searchable {
    public func search(_ string: String) -> SearchResult? {
        return nil
    }
}

extension Int: Searchable {
    public func search(_ string: String) -> SearchResult? {
        guard "\(self)".contains(string) else { return nil }
        return .content(description)
    }
}

extension SendableWrapper: Searchable where WrappedValue: Searchable {
    public func search(_ string: String) -> SearchResult? {
        wrappedValue.search(string)
    }
}
